//! # Implementación de iteradores
//! Los iteradores nos permiten trabajar con colecciones y streams de datos
//! de forma eficiente y prolija. Implementar un iterador es simplemente implementar
//! el trait `Iterator`, y nos provee una gran cantidad de funciones que nos ayudan a
//! darle un enfoque más funcional a nuestro código.

#[derive(Clone)]
/// Nodo de un grafo
/// El ID es para visualizarlo
pub struct Node {
    id: usize, // Sólo para debug
    neighbors: Vec<usize>
}

/// Grafo: posee un vector de nodos
pub struct Graph {
    nodes: Vec<Node>
}

/// Iterador del grafo Depth first search
/// Notar que estamos usando un especificador de lifetime
/// Esto se debe a que queremos que nuestro iterador tenga su 
/// largo de vida acotado por el grafo que itera.
/// Esto se debe a que almacenamos una referencia en vez de tomar ownership del grafo
pub struct DfsIterator<'a> {
    graph: &'a Graph,
    visited: Vec<usize>,
    queue: Vec<usize>
}

impl Node {
    /// Crea un nodo con un id y sin aristas
    pub fn new(id: usize) -> Self {
        let neighbors = Vec::new();
        Node {id, neighbors}
    }

    /// Agrega un arista
    pub fn add_dest(&mut self, dest: usize) {
        self.neighbors.push(dest);
    }

    /// Es "Hoja" si no tiene aristas
    pub fn leaf(&self) -> bool {
        self.neighbors.len() == 0
    }
}

impl Graph {
    /// Crea un grafo para N nodos
    pub fn new(n_nodes: usize) -> Self {
        let mut nodes = Vec::with_capacity(n_nodes);
        for i in 0..n_nodes {
            nodes.push(Node::new(i));
        }
        Graph {nodes}
    }

    /// Agrega un arista al grafo
    pub fn connect(&mut self, src: usize, dest: usize) {
        let node = &mut self.nodes[src];
        node.add_dest(dest);
    }

    /// Obtiene el primer nodo que no es hoja
    /// (No es un buen criterio, pero se puede ver el uso de `position`)
    pub fn get_root(&self) -> Option<usize> {
        self.nodes.iter().position(|node| !node.leaf())
    }

    /// Obtiene los índices de los hijos de un nodo
    pub fn get_children(&self, idx: usize) -> Vec<usize> {
        self.nodes[idx].neighbors.clone()
    }

    /// Crea un iterador Dfs
    pub fn iter(&self) -> DfsIterator {
        DfsIterator::new(&self)
    }
}


// Traits para imprimir en pantalla
// Implementamos el trait std::fmt::Display para poder usar macros de escritura cómodamente
impl std::fmt::Display for Node {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "Nodo {}: {:?}", self.id, self.neighbors)
    }
}


impl std::fmt::Display for Graph {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        for node in &self.nodes {
            write!(f, "{}\n", node)?;
        }
        Ok(())
    }
}

// El iterador Dfs
//
// No es exactamente un iterador DFS pero a fines didácticos sirve <3
impl<'a> DfsIterator<'a> {
    /// Crea un iterador DFS nuevo
    /// Arranca con un vector de nodos visitados, y encola
    /// la "raiz" de nuestro grafo (nuevamente, hay muchas mejoras por aplicar)
    pub fn new(graph: &'a Graph) -> Self {
        let visited = vec![];
        let mut queue = vec![];
        if let Some(root) = graph.get_root() {
            queue.push(root);
        }
        Self {visited, queue, graph}
    }
}

impl Iterator for DfsIterator<'_> {
    /// Especifico el tipo de dato que voy a devolver al iterar
    type Item = usize;

    /// Función necesaria para implementar el trait `Iterator`
    /// Debo devolver Some(elemento) mientras tenga un elemento para recorrer
    /// Si termino de iterar, devuelvo None
    fn next(&mut self) -> Option<<Self as Iterator>::Item> {
        let next_val = self.queue.pop();
        if let Some(val) = next_val {
            if !self.visited.contains(&val) {
                self.queue.append(&mut self.graph.get_children(val));
                self.visited.push(val);
            } else {
                return self.next();
            }
        }
        next_val
    }
}