//! # Uso de iteradores en conjunto de canales
//! 
//! Muchas veces utilizamos hilos para procesar streams de datos. En este ejemplo
//! creamos múltiples streams de productores y los procesamos en hilos separados
//! Estos hilos encolan datos para un receiver en común, y el hilo "main" consume esos datos
use std::iter;

use std::sync::mpsc::{channel, Sender};
use std::thread;


struct MessageProducer {
    msg: String,
    times: usize
}

impl MessageProducer {
    pub fn new(msg_str: &str, times: usize) -> Self {
        Self {msg: msg_str.to_owned(), times}
    }

    pub fn create_messages(&self) -> impl Iterator<Item=String> {
        iter::repeat(self.msg.clone()).take(self.times)
    }

    pub fn send_messages(&self, sender: &Sender<String>) {
        let messages = self.create_messages();
        messages.for_each(|msg| {
            let sleep_time = std::time::Duration::from_millis(20);
            thread::sleep(sleep_time);
            sender.send(msg).unwrap()
        });
    }
}

pub fn run_mpsc() {

    let (sender, receiver) = channel();

    let combos = vec![
        ("Holis", 5),
        ("Lalala", 6),
        ("lelele", 5)
    ];

    // Creamos hilos usando un iterador
    let handlers : Vec<thread::JoinHandle<()>> = combos.iter().map(|conf| {
        // Notar que el sender se debe clonar ¿Por qué? ¿Qué pasa si lo clono dentro del thread?
        let sender_local = sender.clone();
        // ¿Qué pasa si creo este producer dentro del thread?
        let producer = MessageProducer::new(conf.0, conf.1);
        thread::spawn(move || {
           producer.send_messages(&sender_local);
        })
    // Realizo un collect para consumir el iterador y guardar el contenido en un vector
    // ¿Qué pasa si no recolecto los handlers? ¿Cuándo se evalua el map?
    }).collect();

    // ¿Por qué debo hacer drop del sender?
    drop(sender);

    for msg in receiver.iter().take(5) {
        println!("{:?}", msg);
    }

    for handle in handlers {
        handle.join().unwrap();
    }
}

